﻿///
/// @file
/// @brief C to XMI Converter
/// Entry point
///
/// @author Chiharu
///
/// Copyright (C) 2012-2012 Atelier of Chiharu
///

#include <fstream>
#include <iostream>
#include <exception>

#include "cparser/Lexer.hpp"

namespace {

///
/// 使用方法を表示する。
///
void usage()
{
	std::cout << "usage: " << std::endl;
	std::cout << "  > ctoxmi {input}" << std::endl;
	std::cout << "    {input} : input file name" << std::endl;
}

};

///
/// エントリ ポイント
///
/// @param[in] argc コマンド ライン引数の数を指定する。
/// @param[in] argv コマンド ライン引数を指定する。
///
/// @return 常に 0 を返す。
///
int main(int argc, char** argv)
{
	if (argc < 2) {
		usage();
		return 0;
	}
	
	try {
		std::ifstream aInput(argv[1]);
		
		enum {
			
			// コメント
			tokenComment = 0xff,
			tokenCommentCpp,
			
			// キーワード
			tokenAuto,
			tokenBool,
			tokenBreak,
			tokenCase,
			tokenChar,
			tokenComplex,
			tokenConst,
			tokenContinue,
			tokenDefault,
			tokenDo,
			tokenDouble,
			tokenElse,
			tokenEnum,
			tokenExtern,
			tokenFloat,
			tokenFor,
			tokenGoto,
			tokenIf,
			tokenImaginary,
			tokenInline,
			tokenInt,
			tokenLong,
			tokenRegister,
			tokenRestrict,
			tokenReturn,
			tokenShort,
			tokenSigned,
			tokenSizeof,
			tokenStatic,
			tokenStruct,
			tokenSwitch,
			tokenTypedef,
			tokenUnion,
			tokenUnsigned,
			tokenVoid,
			tokenVolatile,
			tokenWhile,
			
			// シンボル
			tokenSymbol,
			
			// 定数（整数）
			tokenConstant,
			
			// 定数（実数）
			tokenConstantReal,
			
			// 文字列
			tokenLiteral,
			
			tokenEllipsis,
			tokenRightShiftAssign,
			tokenLeftShiftAssign,
			tokenAddAssign,
			tokenSubAssign,
			tokenMulAssign,
			tokenDivAssign,
			tokenModAssign,
			tokenAndAssign,
			tokenXorAssign,
			tokenOrAssign,
			tokenRightShift,
			tokenLeftShift,
			tokenInc,
			tokenDec,
			tokenArrow,
			tokenCondAnd,
			tokenCondOr,
			tokenCondLE,
			tokenCondGE,
			tokenCondEQ,
			tokenCondNE,
		//	';',
		//	'{',
		//	'}',
		//	',',
		//	':',
		//	'=',
		//	'(',
		//	')',
		//	'[',
		//	']',
		//	'.',
		//	'&',
		//	'!',
		//	'~',
		//	'-',
		//	'+',
		//	'*',
		//	'/',
		//	'%',
		//	'<',
		//	'>',
		//	'^',
		//	'|',
		//	'?',
			
		};
		
		const cparser::LexerItem aItems[] = {
			// コメント
			{ cparser::lexerItemMatch, "/*",          tokenComment    },
			{ cparser::lexerItemRegex, "//[^\n]*", tokenCommentCpp },
			
			// キーワード
			{ cparser::lexerItemMatch, "auto",       tokenAuto      },
			{ cparser::lexerItemMatch, "_Bool",      tokenBool      },
			{ cparser::lexerItemMatch, "break",      tokenBreak     },
			{ cparser::lexerItemMatch, "case",       tokenCase      },
			{ cparser::lexerItemMatch, "char",       tokenChar      },
			{ cparser::lexerItemMatch, "_Complex",   tokenComplex   },
			{ cparser::lexerItemMatch, "const",      tokenConst     },
			{ cparser::lexerItemMatch, "continue",   tokenContinue  },
			{ cparser::lexerItemMatch, "default",    tokenDefault   },
			{ cparser::lexerItemMatch, "do",         tokenDo        },
			{ cparser::lexerItemMatch, "double",     tokenDouble    },
			{ cparser::lexerItemMatch, "else",       tokenElse      },
			{ cparser::lexerItemMatch, "enum",       tokenEnum      },
			{ cparser::lexerItemMatch, "extern",     tokenExtern    },
			{ cparser::lexerItemMatch, "float",      tokenFloat     },
			{ cparser::lexerItemMatch, "for",        tokenFor       },
			{ cparser::lexerItemMatch, "goto",       tokenGoto      },
			{ cparser::lexerItemMatch, "if",         tokenIf        },
			{ cparser::lexerItemMatch, "_Imaginary", tokenImaginary },
			{ cparser::lexerItemMatch, "inline",     tokenInline    },
			{ cparser::lexerItemMatch, "int",        tokenInt       },
			{ cparser::lexerItemMatch, "long",       tokenLong      },
			{ cparser::lexerItemMatch, "register",   tokenRegister  },
			{ cparser::lexerItemMatch, "restrict",   tokenRestrict  },
			{ cparser::lexerItemMatch, "return",     tokenReturn    },
			{ cparser::lexerItemMatch, "short",      tokenShort     },
			{ cparser::lexerItemMatch, "signed",     tokenSigned    },
			{ cparser::lexerItemMatch, "sizeof",     tokenSizeof    },
			{ cparser::lexerItemMatch, "static",     tokenStatic    },
			{ cparser::lexerItemMatch, "struct",     tokenStruct    },
			{ cparser::lexerItemMatch, "switch",     tokenSwitch    },
			{ cparser::lexerItemMatch, "typedef",    tokenTypedef   },
			{ cparser::lexerItemMatch, "union",      tokenUnion     },
			{ cparser::lexerItemMatch, "unsigned",   tokenUnsigned  },
			{ cparser::lexerItemMatch, "void",       tokenVoid      },
			{ cparser::lexerItemMatch, "volatile",   tokenVolatile  },
			{ cparser::lexerItemMatch, "while",      tokenWhile     },
			
			{ cparser::lexerItemRegex, "[a-zA-Z_]([a-zA-Z_]|[0-9])*", tokenSymbol },
			
			// 定数（整数）
			{ cparser::lexerItemRegex, "0[xX][a-fA-F0-9]+((u|U)|(u|U)?(l|L|ll|LL)|(l|L|ll|LL)(u|U))?", tokenConstant },
			{ cparser::lexerItemRegex, "0[0-7]*((u|U)|(u|U)?(l|L|ll|LL)|(l|L|ll|LL)(u|U))?",           tokenConstant },
			{ cparser::lexerItemRegex, "[1-9][0-9]*((u|U)|(u|U)?(l|L|ll|LL)|(l|L|ll|LL)(u|U))?",       tokenConstant },
			{ cparser::lexerItemRegex, "L?'(\\\\.|[^\\\\'\\n])+'",                                     tokenConstant },
			
			// 定数（実数）
			{ cparser::lexerItemRegex, "[0-9]+([Ee][+-]?[0-9]+)(f|F|l|L)?", tokenConstant },
			{ cparser::lexerItemRegex, "[0-9]*\\.[0-9]+([Ee][+-]?[0-9]+)?(f|F|l|L)?", tokenConstant },
			{ cparser::lexerItemRegex, "[0-9]+\\.[0-9]+([Ee][+-]?[0-9]+)?(f|F|l|L)?", tokenConstant },
			{ cparser::lexerItemRegex, "0[xX][a-fA-F0-9]+([Pp][+-]?[0-9]+)(f|F|l|L)?", tokenConstant },
			{ cparser::lexerItemRegex, "0[xX][a-fA-F0-9]*\\.[a-fA-F0-9]+([Pp][+-]?[0-9]+)?(f|F|l|L)?", tokenConstant },
			{ cparser::lexerItemRegex, "0[xX][a-fA-F0-9]+\\.[a-fA-F0-9]*([Pp][+-]?[0-9]+)?(f|F|l|L)?", tokenConstant },
			
			// 文字列
			{ cparser::lexerItemRegex, "L?\\\"(\\\\.|[^\\\\\"\\n])*\\\"", tokenLiteral },
			
			// 記号
			{ cparser::lexerItemMatch, "...", tokenEllipsis         },
			{ cparser::lexerItemMatch, ">>=", tokenRightShiftAssign },
			{ cparser::lexerItemMatch, "<<=", tokenLeftShiftAssign  },
			{ cparser::lexerItemMatch, "+=",  tokenAddAssign        },
			{ cparser::lexerItemMatch, "-=",  tokenSubAssign        },
			{ cparser::lexerItemMatch, "*=",  tokenMulAssign        },
			{ cparser::lexerItemMatch, "/=",  tokenDivAssign        },
			{ cparser::lexerItemMatch, "%=",  tokenModAssign        },
			{ cparser::lexerItemMatch, "&=",  tokenAndAssign        },
			{ cparser::lexerItemMatch, "^=",  tokenXorAssign        },
			{ cparser::lexerItemMatch, "|=",  tokenOrAssign         },
			{ cparser::lexerItemMatch, ">>",  tokenRightShift       },
			{ cparser::lexerItemMatch, "<<",  tokenLeftShift        },
			{ cparser::lexerItemMatch, "++",  tokenInc              },
			{ cparser::lexerItemMatch, "--",  tokenDec              },
			{ cparser::lexerItemMatch, "->",  tokenArrow            },
			{ cparser::lexerItemMatch, "&&",  tokenCondAnd          },
			{ cparser::lexerItemMatch, "||",  tokenCondOr           },
			{ cparser::lexerItemMatch, "<=",  tokenCondLE           },
			{ cparser::lexerItemMatch, ">=",  tokenCondGE           },
			{ cparser::lexerItemMatch, "==",  tokenCondEQ           },
			{ cparser::lexerItemMatch, "!=",  tokenCondNE           },
			
			{ cparser::lexerItemMatch, ";",  ':' },
			{ cparser::lexerItemMatch, "{",  '{' },
			{ cparser::lexerItemMatch, "<%", '{' },
			{ cparser::lexerItemMatch, "}",  '}' },
			{ cparser::lexerItemMatch, "%>", '}' },
			{ cparser::lexerItemMatch, ",",  ',' },
			{ cparser::lexerItemMatch, ":",  ':' },
			{ cparser::lexerItemMatch, "=",  '=' },
			{ cparser::lexerItemMatch, "(",  '(' },
			{ cparser::lexerItemMatch, ")",  ')' },
			{ cparser::lexerItemMatch, "[",  '[' },
			{ cparser::lexerItemMatch, "<:", '[' },
			{ cparser::lexerItemMatch, "]",  ']' },
			{ cparser::lexerItemMatch, ":>", ']' },
			{ cparser::lexerItemMatch, ".",  '.' },
			{ cparser::lexerItemMatch, "&",  '&' },
			{ cparser::lexerItemMatch, "!",  '!' },
			{ cparser::lexerItemMatch, "~",  '~' },
			{ cparser::lexerItemMatch, "-",  '-' },
			{ cparser::lexerItemMatch, "+",  '+' },
			{ cparser::lexerItemMatch, "*",  '*' },
			{ cparser::lexerItemMatch, "/",  '/' },
			{ cparser::lexerItemMatch, "%",  '%'},
			{ cparser::lexerItemMatch, "<",  '<' },
			{ cparser::lexerItemMatch, ">",  '>' },
			{ cparser::lexerItemMatch, "^",  '^' },
			{ cparser::lexerItemMatch, "|",  '|' },
			{ cparser::lexerItemMatch, "?",  '?' },
			
			// 空白記号
			{ cparser::lexerItemRegex, "[ \\t\\v\\n\\f]", cparser::LexerItem::codeSkip },
		};
		cparser::Lexer aLexer(aInput, aItems, aItems + sizeof(aItems) / sizeof(aItems[0]));
		
		while (auto aToken = aLexer.getNextToken()) {
			std::cout << "(" << aToken->code << ") : " << std::string(aToken->begin, aToken->end) << std::endl;
		}
		
	} catch (std::exception& iEx) {
		std::cout << "exception : " << iEx.what() << std::endl;
	}
	
	return 0;
}
