///
/// @file
/// @brief cparser::Lexer
///
/// @author Chiharu
///
/// Copyright (C) 2012-2012 Atelier of Chiharu
///

#include "LexerException.hpp"
#include "Lexer.hpp"
#include <regex>
#include <algorithm>
#include <cassert>

namespace cparser {

namespace {

///
/// g[NNAB
///
/// @param[out] oToken g[Nw肷B
///
inline void clearToken(Token& oToken)
{
	oToken.begin = nullptr;
	oToken.end = nullptr;
	oToken.code = LexerItem::codeSkip;
}

///
/// g[Nݒ肷B
///
/// @param[out] oToken g[Nw肷B
/// @param[out] iBegin g[N̐擪w肷B
/// @param[out] iEnd g[N̖w肷B
/// @param[out] iCode o̒ʒmR[hw肷B
///
inline void resetToken(Token& oToken, const char* iBegin, const char* iEnd, const std::uint32_t iCode)
{
	oToken.begin = iBegin;
	oToken.end = iEnd;
	oToken.code = iCode;
}

///
/// g[NXVB
///
/// @param[in,out] ioToken g[Nw肷B
/// @param[out] iBegin g[N̐擪w肷B
/// @param[out] iEnd g[N̖w肷B
/// @param[out] iCode o̒ʒmR[hw肷B
///
inline void updateToken(Token& ioToken, const char* iBegin, const char* iEnd, const std::uint32_t iCode)
{
	if ((iEnd - iBegin) > (ioToken.end - ioToken.begin)) {
		resetToken(ioToken, iBegin, iEnd, iCode);
	}
}

};

//
// RXgN^
//
Lexer::Lexer(std::istream& iSource, const LexerItem* iBegin, const LexerItem* iEnd):
	mSource(iSource), mItems(iBegin, iEnd), mLine(0), mColumn(0)
{
	// g[ÑNA
	clearToken(mLastToken);
}

//
// fXgN^
//
Lexer::~Lexer()	// nothrow
{
	// do nothing.
}

//
// g[N͂B
//
const Token* Lexer::getNextToken()
{
	clearToken(mLastToken);
	
	while (!isEndOfStream()) {
		// ǂݍ
		if (isEndOfLine()) {
			readNextLine();
		}
		if (isEndOfLine()) {
			continue;
		}
		
		// w蕶̒T
		std::for_each(mItems.begin(), mItems.end(), [this](LexerItem& iItem) {
			if (iItem.type == lexerItemRegex) {
				std::regex aRegex(iItem.pattern);
				std::cmatch aMatch;
				const auto aMatched = std::regex_search(mLineBuffer.data() + mColumn, mLineBuffer.data() + mLineBuffer.length(), aMatch, aRegex);
				if (aMatched && ((mLineBuffer.data() + mColumn) == aMatch[0].first)) {
					updateToken(mLastToken, aMatch[0].first, aMatch[0].second, iItem.code);
				}
			} else {
				const auto aPattenLength = std::strlen(iItem.pattern);
				const auto aMatched = (mLineBuffer.compare(mColumn, aPattenLength, iItem.pattern) == 0);
				if (aMatched) {
					updateToken(mLastToken, iItem.pattern, iItem.pattern + aPattenLength, iItem.code);
				}
			}
		});
		
		if ((mLastToken.begin == nullptr) || (mLastToken.end == nullptr)) {
			throw LexerException(mLine, mColumn);
		}
		
		// g[N̒ʒm
		mColumn += mLastToken.end - mLastToken.begin;
		if (mLastToken.code != LexerItem::codeSkip) {
			return &mLastToken;
		}
		
		clearToken(mLastToken);
	}
	
	return nullptr;
}

//
// w肳ꂽ܂œ̓Xg[XLbvB
//
void Lexer::skip(const char* iUntil)
{
	assert(iUntil != nullptr);
	
	// XLbv
	auto aPos = std::string::npos;
	
	do {
		// ǂݍ
		if (isEndOfStream()) {
			throw LexerException(mLine, mColumn);
		} else if (isEndOfLine()) {
			readNextLine();
		}
		
		// w蕶̒T
		aPos = mLineBuffer.find(iUntil);
		if (aPos == std::string::npos) {
			mColumn = mLineBuffer.size();
		} else {
			mColumn = aPos;
		}
	} while (aPos == std::string::npos);
	
	// g[ÑNA
	clearToken(mLastToken);
}

//
// ̓Xg[̏I[𔻕ʂB
//
const bool Lexer::isEndOfStream()
{
	return isEndOfLine() && mSource.eof();
}

//
// ̓Xg[̍s𔻕ʂB
//
const bool Lexer::isEndOfLine()
{
	return mColumn == mLineBuffer.size();
}

//
// CǂݍށB
//
void Lexer::readNextLine()
{
	if (mSource.eof()) {
		return;
	}
	
	mLineBuffer.clear();
	std::getline(mSource, mLineBuffer);
	mColumn = 0;
	++mLine;
}

};
